require 'uu/os/qos/qos_handler'
require 'uu/os/rest/binary_value'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/property/property_attributes'
require 'uu/os/property/property_entry_attributes'
require 'uu/os/property/property_create'
require 'uu/os/property/property_get_property_list'
require 'uu/os/property/property_set_attributes'
require 'uu/os/property/property_create_collection'
require 'uu/os/property/property_collection_attributes'
require 'uu/os/property/property_set_collection_attributes'
require 'uu/os/property/property_delete_collection'
require 'uu/os/property/property_get_entry_list'
require 'uu/os/property/property_move_collection'
require 'uu/os/property/property_move'
require 'uu/os/property/property_type'
require 'uu/os/attachment/stream_wrapper'
require 'date'

module UU
  module OS
    # UES Property service.
    module Property
      extend self

      # Service path
      PATH = 'ues/core/property/UESProperty'
      # Returns attributes representing the artifact property specified by the property_uri attribute.
      #
      # @param [String, UU::OS::UESURI] property_uri UESURI of Property
      #
      # @return [PropertyAttributes] DTO with artifact property attributes
      def get_attributes(property_uri)
        svc = UU::OS::REST::RemoteClient.new(Property)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', property_uri)
          return Property::PropertyAttributes.new(res)
        end
      end

      # This command sets a new value on the specified property. The data type of value must match the data
      # type of the specified property. If the property value is set as file, is highly recommended to open file
      # in binary mode, for example: File.open("file_name.txt", "rb").
      #
      # @param [String, UU::OS::UESURI] property_uri UES URI of property
      # @param [Object] value The new value of property. Possible types of property's value (there may be additional
      #   business restrictions based on property type):
      #   * Nil value
      #   * Numeric value
      #   * Time value
      #   * String
      #   * UESURI
      #   * BinaryValue
      #   * File
      def set_value(property_uri, value = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = {}
        
        if (value.kind_of?IO)
          value = UU::OS::REST::BinaryValue.new(value)
        end
        if ((value.kind_of?UU::OS::REST::BinaryValue) && (value.data.nil?) && (!value.name.nil? && !value.name.empty?))
          payload[:defaultName] = value.name
        elsif (value.kind_of?UU::OS::REST::BinaryValue) && (!value.data.nil?)
          if (value.data.kind_of?File) && (File.basename(value.data.path) == value.name)
            payload[:defaultName] = value.name
          end
          if (value.data.kind_of?File)
            payload[:value] = create_file_with_file_name(value)
          else
            payload[:value] = create_stream_with_file_name(value)
          end
        else
          payload = value.to_json
        end
         
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('setValue', property_uri, payload)
        end
      end

      # Returns a value of the specified property.
      #
      # @param [String, UU::OS::UESURI] property_uri UESURI of the property
      #
      # @return [Object] The value of the specified property
      def get_value(property_uri)
        svc = UU::OS::REST::RemoteClient.new(Property)
        UU::OS::QoS::QoSHandler.auto_retry do
          raw = svc.raw_get_binary('getValue', property_uri)
          type = raw[2][:uu_property_value_type]
          value = svc.process_result(raw)
          result = deserialize_by_type(value,type)
          if (result.kind_of?UU::OS::REST::BinaryValue)
            if (raw[2][:content_disposition] != nil)
              disps = raw[2][:content_disposition].split(";")
              disps.each do |part|
                kv = part.lstrip.split("=")
                if kv[0]=="filename"
                  result.name = kv[1]
                end
              end
            end
          end
          return result
        end
      end

      # Creates a new property on the specified artifact or in the specified property collection.
      # At the very least, property template or property type and name have to be specified in this command.
      # Value of the property is optional.
      #
      # @param [String, UU::OS::UESURI] parent_node_uri UESURI of the artifact or of the property collection where the new property is to be created
      # @param [PropertyCreate] property DTO containing attributes of a new property
      def create(parent_node_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        prop = UU::OS::Property::PropertyCreate.new(property)
        value = prop.value
        
        if (value.kind_of?File)
          prop.value = UU::OS::REST::BinaryValue.new(value)
        end
        if (prop.value.kind_of?UU::OS::REST::BinaryValue)
          payload = prop.to_hash
          payload.delete(:value)
          if (prop.value.data.kind_of?File)
            payload[:value] = create_file_with_file_name(prop.value)
          else
            payload[:value] = create_stream_with_file_name(prop.value)
          end
        else
          payload = prop.to_json
        end
     
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', parent_node_uri, payload)
        return UU::OS::UESURI.new(res)
        end
      end
     

      # Deletes the specified property.
      #
      # @param [String, UU::OS::UESURI] property_uri UESURI of the property which is about to be deleted.
      def delete(property_uri)
        svc = UU::OS::REST::RemoteClient.new(Property)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', property_uri)
        end
      end

      # Sets attributes of the specified property.
      #
      #
      # @param [String, UU::OS::UESURI] property_uri UESURI of the property whose attributes are about to be changed.
      # @param [PropertySetAttributes] property DTO containing the attributes of the property to be set.
      #
      # @return [UU::OS::UESURI] UESURI of the modified property.
      def set_attributes(property_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertySetAttributes.new(property).to_json
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', property_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a list of entries (properties or property collections) enlisted in the specified
      # property collection or on the specified artifact (properties from the artiffact root property collection
      # are returned in that case). Command does not return all entries, only a sublist. The list of returned
      # entries could be filtered by the name, code or type of an entry. The list is sorted by the name (and codes
      # in case are equal) of a entry by default, or it can be sorted arbitrarily by any of entry attributes usable
      # for the filtering or these attributes combinations. See the UES query documentation for more information.
      #
      # @param [String, UU::OS::UESURI] parent_node_uri UESURI of the artifact or of the property collection from where entries (properties
      #   or property collections) are listed.
      # @param [PropertyGetEntryList] criteria DTO containing criteria for listing entries.
      #
      # @return [UU::OS::REST::ResultList<Property::PropertyGetEntryList, Property::PropertyEntryAttributes>] List of DTO representing property entries (properties or property collections)
      def get_entry_list(parent_node_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        dto = Property::PropertyGetEntryList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        svc.add_parameter('recursive', dto.recursive)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getEntryList', parent_node_uri)
          res = UU::OS::REST::ResultList.new(Property::PropertyGetEntryList, Property::PropertyEntryAttributes, res)
          return res
        end
      end

      # This command creates a new property collection on the specified artifact or in
      # the specified property collection. At the very least property collection name
      # has to be specified in this command.
      #
      # @param [String, UU::OS::UESURI] parent_node_uri UESURI of the artifact or of the property collection where the new property collection is created.
      # @param [PropertyCreateCollection] property DTO containing attributes of the new property collection.
      #
      # @return [UU::OS::UESURI] UESURI of the new property collection.
      def create_collection(parent_node_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertyCreateCollection.new(property).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('createCollection', parent_node_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # This command returns attributes representing the property collection specified
      # by the collectionUri parameter. The command does not change the state of the property
      # collection in the system.
      #
      # @param [String, UU::OS::UESURI] collection_uri UESURI of the property collection which attributes are returned.
      #
      # @return [PropertyCollectionAttributes] DTO containing the property collection attributes
      def get_collection_attributes(collection_uri)
        svc = UU::OS::REST::RemoteClient.new(Property)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getCollectionAttributes', collection_uri)
          return Property::PropertyCollectionAttributes.new(res)
        end
      end

      # This command sets attributes of the specified property collection.
      #
      # @param [String, UU::OS::UESURI] collection_uri UESURI of the property collection whose attributes are about to be changed.
      # @param [PropertySetCollectionAttributes] property DTO containing the attributes of the property collection to be set.
      #
      # @return [UU::OS::UESURI] UESURI of the property collection whose attributes are to changed.
      def set_collection_attributes(collection_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertySetCollectionAttributes.new(property).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setCollectionAttributes', collection_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # This command deletes the specified property collection from the system. When
      # the collection is not empty, it's need explicite to allow deletion of the collection
      # with its all nested objects.
      #
      # @param [String, UU::OS::UESURI] collection_uri UESURI of the property collection which is about to be deleted
      # @param [PropertyDeleteCollection] property DTO containing settings for deletion
      def delete_collection(collection_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertyDeleteCollection.new(property).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('deleteCollection', collection_uri, payload)
        end
      end

      # This command moves a property collection to the new location in the same artifact.
      # This collection can not be base collection. New location can be property collection
      # or artifact, and it can not be child of property collection which is about to move.
      # New location can not be child of the moved collection. User must have access rights
      # to modify property collection on artifact to which is collection related.
      #
      # @param [String, UU::OS::UESURI] collection_uri UESURI of the property collection which is about to be moved.
      # @param [PropertyMoveCollection] property DTO contains parameters that specify the move
      def move_collection(collection_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertyMoveCollection.new(property).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('moveCollection', collection_uri, payload)
        end
      end

      # Moves the specified property to a new location in same artifact in which the property
      # is.
      # For the property move - the logged user, which invokes this command,
      # must have access rights to modify target property collection.
      #
      # @param [String, UU::OS::UESURI] property_uri UESURI of the property which is about to be moved.
      # @param [PropertyMove] property DTO containing attributes specifying a location to where the property is about to be moved.
      def move(property_uri, property = nil)
        svc = UU::OS::REST::RemoteClient.new(Property)
        payload = UU::OS::Property::PropertyMove.new(property).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('move', property_uri, payload)
        end
      end

      private

      def deserialize_by_type(value,type)

        return nil unless value
        result = nil
        case type
        when PropertyType::BINARY, PropertyType::IMAGE
            result = UU::OS::REST::BinaryValue.new(value, true)
        when PropertyType::DATE
          if value != nil and value != ""
            result = DateTime.strptime(eval(String.new("\"" + value.read + "\"")),'%Y-%m-%d')
          end
        when PropertyType::INTERTERRITORIAL_REFERENCE,PropertyType::REFERENCE
          if value != nil and value !=""
            result = UU::OS::UESURI.new(eval(String.new("\"" + value.read + "\"")))
          end
        when PropertyType::NUMBER
            result = eval(String.new("\"" + value.read + "\"")).to_f
        when PropertyType::SHORT_TEXT, PropertyType::TEXT, PropertyType::URL
            result = String.new(value.read).force_encoding("UTF-8")
        end
        return result

      end
      
      def create_stream_with_file_name(bin_val)
        bin_val.data.rewind
        UU::OS::Attachment::StreamWrapper.new(bin_val.data, "#{bin_val.name}")
      end
      
      def create_file_with_file_name(bin_val)
        file = File.new(bin_val.data.path,'rb')
        path = bin_val.name.nil? || bin_val.name.strip.empty? ? file.path : "#{bin_val.name}"
        UU::OS::Attachment::StreamWrapper.new(file, path)
      end

    end
  end
end
